!!----
!!---- MODULE: Iso_Surfaces
!!----   INFO: Iso_Surfaces Representation
!!----         (Marching_Cubes & 2D Contour)
!!----
!!----         Copyright(C) 1999-2003,              Version: 1.0
!!----         Juan Rodriguez-Carvajal & Javier Gonzalez-Platas
!!----
!!---- HISTORY
!!----    Update: February - 2003
!!----
!!----    Nov - 1999: Created by JGP
!!----
!!---- VARIABLES
!!----    CUBE_INFO_TYPE
!!----    CUBE_INFO
!!----    MAX_POINTS
!!----
!!---- PROCEDURES
!!----    Subroutines:
!!----       CALCULATE_CONTOUR2D
!!----       CALCULATE_MESH
!!----       LOAD_EXTENDEDMAP
!!----
!!
 Module Iso_Surfaces
    !---- Variables ----!
    implicit none

    private

    !---- List of public subroutines ----!
!    public :: Calculate_Contour2D, Calculate_Mesh, Load_ExtendedMap, Set_Cube_Info, Load_Section
    public :: Calculate_Contour2D, Load_Section

    !---- Definitions ----!
    real, parameter :: eps=0.0001

    !!----
    !!----  TYPE :: CUBE_INFO_TYPE
    !!--..
    !!----  Type :: Cube_Info_Type
    !!----     integer                :: NElem     ! Number of Elemens
    !!----     integer                :: Code      ! Code of Elements
    !!----                                           1:               4:
    !!----                                           2:               5:
    !!----                                           3:               6:
    !!----     integer, dimension(12) :: Edges     ! Code for Edge connections
    !!----  End Type Cube_Info_Type
    !!----
    !!----  Update: February - 2003
    !!
    Type, public :: Cube_Info_Type
       integer                :: NElem
       integer                :: Code
       integer, dimension(12) :: Edges
    End Type Cube_Info_Type

    !!----
    !!---- CUBE_INFO
    !!----     Type (Cube_Info_Type), dimension(0:255) :: Cube_Info
    !!----
    !!----     Information of Mesh in a cube
    !!----
    !!---- Update: February - 2003
    !!
    Type (Cube_Info_Type), dimension(0:255), public :: Cube_Info

    !!----
    !!---- MAX_POINTS
    !!----    integer, parameter, public ::  Max_Points
    !!----
    !!----    Number of maximum points permitted
    !!----
    !!---- Update: February - 2003
    !!
!   integer, parameter, public ::  Max_Points    = 379980
!    integer, parameter, public ::  Max_Points    = 379972

 Contains

    !---- Subroutines ----!

    !!----
    !!---- Subroutine Calculate_Contour2D(d,ilb,iub,jlb,jub,x,y,z,nlv,ntp,,xyz,Max_Points,ier)
    !!----    integer,                           intent(in)     :: ilb,iub
    !!----    integer,                           intent(in)     :: jlb,jub
    !!----    real, dimension (ilb:iub,jlb:jub), intent(in)     :: d
    !!----    real, dimension (ilb:iub),         intent(in)     :: x
    !!----    real, dimension (jlb:jub),         intent(in)     :: y
    !!----    real, dimension (:),               intent(in)     :: z
    !!----    integer,                           intent(in)     :: nlv ! numeri de liveli
    !!----    integer,                           intent(in out) :: ntp ! punti
    !!----    real, dimension (:,:),             intent(in out) :: xyz
    !!----       integer,                        intent(in)     :: Max_Points
    !!----    integer,                           intent(in out) :: ier
    !!
    Subroutine Calculate_Contour2D(d,ilb,iub,jlb,jub,x,y,z,nlv,npt,xyz,Max_Points,ier)
       !---- Arguments ----!
       integer,                           intent(in)     :: ilb,iub
       integer,                           intent(in)     :: jlb,jub
       real, dimension (ilb:iub,jlb:jub), intent(in)     :: d
       real, dimension (ilb:iub),         intent(in)     :: x
       real, dimension (jlb:jub),         intent(in)     :: y
       real, dimension (:),               intent(in)     :: z
       integer,                           intent(in)     :: nlv ! numeri de liveli
       integer,                           intent(in out) :: npt ! punti
       real, dimension (:,:),             intent(in out) :: xyz
       integer,                           intent(in)     :: Max_Points
       integer,                           intent(in out) :: ier

       !---- Local variables ----!
       integer                             :: j,i,k,m,m1,m2,m3
       integer                             :: cases, p1, p2
       integer, dimension (0:4)            :: sh
       integer, dimension (1:4)            :: im=(/0,1,1,0/), jm=(/0,0,1,1/)
       integer, dimension (-1:1,-1:1,-1:1) :: castab
       real, dimension (0:4)               :: h, xh, yh
       real                                :: xsect,ysect,dmin,dmax,x1,y1,x2,y2

       !---- Use statement functions for the line intersections
       xsect(p1,p2)=(h(p2)*xh(p1)-h(p1)*xh(p2))/(h(p2)-h(p1))
       ysect(p1,p2)=(h(p2)*yh(p1)-h(p1)*yh(p2))/(h(p2)-h(p1))

       !---- Init ----!
       castab= reshape ( (/0,0,9,0,1,5,7,4,8,0,3,6,2,3,2,6,3,0,8,4,7,5,1,0,9,0,0/),(/3,3,3/) )

       !---- Scan the arrays, top down, left to right within rows
       do j=jub-1,jlb,-1
          do i=ilb,iub-1
             call DoSingleIteration()
             dmin=min(d(i,j),d(i,j+1),d(i+1,j),d(i+1,j+1))
             dmax=max(d(i,j),d(i,j+1),d(i+1,j),d(i+1,j+1))

             if (dmax >= z(1) .and. dmin <= z(nlv)) then
                do k=1,nlv
                   if (z(k) >= dmin .and. z(k) <= dmax) then
                      do m=4,0,-1
                         if (m > 0) then
                            h(m)=d(i+im(m),j+jm(m)) -z(k)
                            xh(m)=x(i+im(m))
                            yh(m)=y(j+jm(m))
                         else
                            h(0)=0.25*(h(1)+h(2)+h(3)+h(4))
                            xh(0)=0.5*(x(i)+x(i+1))
                            yh(0)=0.5*(y(j)+y(j+1))
                         end if
                         if ( h(m) > 0.0) then
                            sh(m)=+1
                         else if ( h(m) < 0.0) then
                            sh(m)=-1
                         else
                            sh(m)=0
                         end if
                      end do

                      !--- Scan each triangle in the box
                      do m=1,4
                         m1=m
                         m2=0
                         if (m /= 4) then
                            m3=m+1
                         else
                            m3=1
                         end if
                         cases=castab(sh(m1),sh(m2),sh(m3))
                         if (cases /= 0) then
                            select case (cases)
                               case (1)
                                  ! Case 1 - Line between vertices 1 and 2
                                  x1=xh(m1)
                                  y1=yh(m1)
                                  x2=xh(m2)
                                  y2=yh(m2)

                               case (2)
                                  ! Case 2 - Line between vertices 2 and 3
                                  x1=xh(m2)
                                  y1=yh(m2)
                                  x2=xh(m3)
                                  y2=yh(m3)

                               case (3)
                                  ! Case 3 - Line between vertices 3 and 1
                                  x1=xh(m3)
                                  y1=yh(m3)
                                  x2=xh(m1)
                                  y2=yh(m1)

                               case (4)
                                  ! Case 4 - Line between vertices 1 and side 2-3
                                  x1=xh(m1)
                                  y1=yh(m1)
                                  x2=xsect(m2,m3)
                                  y2=ysect(m2,m3)

                               case (5)
                                  ! Case 5 - Line between vertices 2 and side 3-1
                                  x1=xh(m2)
                                  y1=yh(m2)
                                  x2=xsect(m3,m1)
                                  y2=ysect(m3,m1)

                               case (6)
                                  ! Case 6 - Line between vertices 3 and side 1-2
                                  x1=xh(m3)
                                  y1=yh(m3)
                                  x2=xsect(m1,m2)
                                  y2=ysect(m1,m2)

                               case (7)
                                  ! Case 7 - Line between sides 1-2 and  2-3
                                  x1=xsect(m1,m2)
                                  y1=ysect(m1,m2)
                                  x2=xsect(m2,m3)
                                  y2=ysect(m2,m3)

                               case (8)
                                  ! Case 8 - Line between sides 2-3 and  3-1
                                  x1=xsect(m2,m3)
                                  y1=ysect(m2,m3)
                                  x2=xsect(m3,m1)
                                  y2=ysect(m3,m1)

                               case (9)
                                  ! Case 9 - Line between sides 3-1 and  1-2
                                  x1=xsect(m3,m1)
                                  y1=ysect(m3,m1)
                                  x2=xsect(m1,m2)
                                  y2=ysect(m1,m2)
                            end select

                            if (npt+2 <= Max_Points) then
                               npt=npt+1
                               xyz(1,npt)=x1
                               xyz(2,npt)=y1
                               xyz(3,npt)=real(k)

                               npt=npt+1
                               xyz(1,npt)=x2
                               xyz(2,npt)=y2
                               xyz(3,npt)=real(k)
                            else
                               ier = -1
                            end if

                         end if
                      end do
                   end if
                end do
             end if
          end do
       end do

       return
    End Subroutine Calculate_Contour2D

   !!---- SUBROUTINE Load_Section(Rho,ngrid,imap,section,limits,ngrid2,dmap)
   !!----    real, dimension(:,:,:), intent(in) :: rho
   !!----    integer, dimension(3),  intent(in) :: ngrid
   !!----    integer,                intent(in) :: imap
   !!----    integer,                intent(in) :: section
   !!----    real, dimension(2,2),   intent(in) :: limits
   !!----    integer, dimension(2),  intent(in) :: ngrid2
   !!----    real, dimension(:,:),   intent(out):: dmap
   !!----
   !!----    This routine only works with fractional coordinates
   !!----
   !!---- Updated: May - 2003
   !!
   Subroutine Load_Section(Rho,ngrid,imap,section,limits,ngrid2,dmap)
      !---- Arguments ----!
      real, dimension(:,:,:), intent(in) :: rho
      integer, dimension(3),  intent(in) :: ngrid
      integer,                intent(in) :: imap
      integer,                intent(in) :: section
      real, dimension(2,2),   intent(in) :: limits
      integer, dimension(2),  intent(in) :: ngrid2
      real, dimension(:,:),   intent(out):: dmap

      !---- Local Variables ----!
      integer :: ndimx, ndimy
      integer :: i,j,ii1,ii2,jj1,jj2,kk

      real    :: uinc,vinc
      real    :: uu,vv,x1,y1
      real    :: dx,dy,f1,f2,f3,z1,z2,z3,z4
      real    :: xx1,xx2,yy1,yy2

      !---- Contorno ----!
      dmap = 0.0

      select case (imap)
         case (1)
            ndimx=ngrid(1)+1
            ndimy=ngrid(2)+1
            dx=1.0/real(ngrid(1))
            dy=1.0/real(ngrid(2))

            kk=mod(section,ngrid(3)+1)
            if (kk == 0) kk=1

         case (2)
            ndimx=ngrid(2)+1
            ndimy=ngrid(3)+1
            dx=1.0/real(ngrid(2))
            dy=1.0/real(ngrid(3))

            kk=mod(section,ngrid(1)+1)
            if (kk == 0) kk=1

         case (3)
            ndimx=ngrid(3)+1
            ndimy=ngrid(1)+1
            dx=1.0/real(ngrid(3))
            dy=1.0/real(ngrid(1))

            kk=mod(section,ngrid(2)+1)
            if (kk == 0) kk=1
      end select

      uinc=(limits(1,2) - limits(1,1))/real(ngrid2(1)-1)
      vinc=(limits(2,2) - limits(2,1))/real(ngrid2(2)-1)

      do i=1,ngrid2(1)
         uu=limits(1,1)+uinc*(i-1)

         do j=1,ngrid2(2)
            vv=limits(2,1)+vinc*(j-1)

            x1=mod(uu+10.0,1.0)
            y1=mod(vv+10.0,1.0)

            !---- Entre que nodos X ----!
            do ii1=1,ndimx-1
               xx1=(ii1-1)*dx
               xx2=xx1+dx
               if (abs(xx1-x1) <= 0.0001) then
                  exit
               else if (x1 > xx1 .and. x1 <= xx2) then
                    exit
               end if
            end do
            ii2=ii1+1
            if (ii2 == ndimx) ii2=1

            !---- Entre que nodos Y ----!
            do jj1=1,ndimy-1
               yy1=(jj1-1)*dy
               yy2=yy1+dy
               if (abs(yy1-y1) <= 0.0001) then
                  exit
               else if (y1 > yy1 .and. y1 <= yy2) then
                  exit
               end if
            end do
            jj2=jj1+1
            if (jj2 == ndimy) jj2=1

            select case (imap)
               case (1)
                  z1=rho(ii1,jj1,kk)
                  z2=rho(ii2,jj1,kk)
                  z3=rho(ii1,jj2,kk)
                  z4=rho(ii2,jj2,kk)
               case (2)
                  z1=rho(kk,ii1,jj1)
                  z2=rho(kk,ii2,jj1)
                  z3=rho(kk,ii1,jj2)
                  z4=rho(kk,ii2,jj2)
               case (3)
                  z1=rho(jj1,kk,ii1)
                  z2=rho(jj1,kk,ii2)
                  z3=rho(jj2,kk,ii1)
                  z4=rho(jj2,kk,ii2)
            end select

            f1=( (x1-xx1)/(xx2-xx1) )*(z2-z1) + z1
            f2=( (y1-yy1)/(yy2-yy1) )*(z3-z1)
            f3=( (x1-xx1)/(xx2-xx1) )*( (y1-yy1)/(yy2-yy1) )*(z1-z2-z3+z4)

            dmap(i,j)=dmap(i,j) + f1+f2+f3

         end do
      end do

      return
   End Subroutine Load_Section

 End Module Iso_Surfaces

