module rinterface
!
!  Procedures to interface refinement with GUI
!
   use atom_basic, only: atom_type
   use elements, only: element_type

   implicit none

   type crystal_t
      character(len=300)       :: cname
      real, dimension(6)       :: cell
      integer, dimension(6)    :: cellcod
      character(len=16)        :: spgname
      real                     :: scal
      integer                  :: scalcod
      integer                  :: kopcorr = 0
      integer, dimension(3)    :: hklpo
      integer                  :: gcode
      real                     :: gpar = 1.0
      logical                  :: riding_model
      real                     :: adp_factor = 1.2
   end type crystal_t

   type dataset_t
     real                      :: zero=0.0
     real                      :: sdisp=0.0
     real                      :: stran=0.0
     integer                   :: zcode,sdcode,stcode
     integer                   :: bcode
   end type dataset_t

   type(crystal_t), dimension(:), allocatable :: cryph
   type(dataset_t), dimension(:), allocatable :: datas

   private :: new_crystal_t, numcryst

contains

   subroutine make_crystal_t()
   use variables, only: cryst
   use elements
   use atom_type_util
   use crystal_phase
   integer :: nph
!
   call new_crystal_t(cryph,numphase(cryst))
   do nph=1,numphase(cryst)
      cryph(nph)%cname = cryst(nph)%cr_name
      cryph(nph)%cell = cryst(nph)%get_cellpar()
      cryph(nph)%cellcod = cryst(nph)%get_cellcod()
      cryph(nph)%spgname = cryst(nph)%spg%symbol_xhm
      cryph(nph)%scal = cryst(nph)%scal
      cryph(nph)%scalcod = cryst(nph)%scode
      cryph(nph)%riding_model = cryst(nph)%riding_model
      cryph(nph)%adp_factor = cryst(nph)%adp_factor
   enddo
!
   end subroutine make_crystal_t

!----------------------------------------------------------------------------------------------------

   subroutine make_dataset_t()
   USE variables, only: dataset
   integer :: nd
   call new_dataset_t(datas,1)
   do nd=1,1
      datas(nd)%zero = dataset(nd)%zero
      datas(nd)%sdisp = dataset(nd)%sdisp
      datas(nd)%stran = dataset(nd)%stran

      datas(nd)%zcode = dataset(nd)%zcode
      datas(nd)%sdcode = dataset(nd)%sdcode
      datas(nd)%stcode = dataset(nd)%stcode

      !datas(nd)%cond = dataset(nd)%cond
      datas(nd)%bcode = dataset(nd)%bcode(1)
   enddo
   end subroutine make_dataset_t

!----------------------------------------------------------------------------------------------------

   subroutine get_info_crystal(numph,natoms,nelem)
   use variables, only: cryst
   use crystal_phase
   use arrayutil
   integer, intent(out)                            :: numph
   integer, dimension(:), allocatable, intent(out) :: natoms,nelem
   integer                                         :: nph
   numph = numphase(cryst)
   call new_array(natoms,numph)
   call new_array(nelem,numph)
   do nph=1,numph
      natoms(nph) = cryst(nph)%natoms()
      nelem(nph) = cryst(nph)%numelem()
   enddo
   end subroutine get_info_crystal

!----------------------------------------------------------------------------------------------------

   subroutine print_crystal_t(kpr)
   use elements
   use atom_type_util
   integer, intent(in) :: kpr
   integer             :: nph
!
   do nph=1,numcryst(cryph)
      write(kpr,'(a)') 'Name: '//trim(cryph(nph)%cname)
      write(kpr,'(a,6f12.5)') 'Cell: ',cryph(nph)%cell(:)
      write(kpr,'(a,6i12)')   'Code: ',cryph(nph)%cellcod(:)
      write(kpr,'(a)') 'Space group: '//trim(cryph(nph)%spgname)
      write(kpr,'(a,f0.3,a,i0)')   'Scale: ',cryph(nph)%scal,' Code: ',cryph(nph)%scalcod
!!mr      call print_elements(cryph(nph)%elem,kpr)
!!mr      call print_atoms(cryph(nph)%at,kpr=kpr)
   enddo
!
   end subroutine print_crystal_t

!----------------------------------------------------------------------------------------------------

   integer function numcryst(phase)
   type(crystal_t), dimension(:), allocatable, intent(in) :: phase
   if (allocated(phase)) then
       numcryst = size(phase)
   else
       numcryst = 0
   endif
   end function numcryst

!----------------------------------------------------------------------------------------------------

   subroutine new_crystal_t(vetr,n)
   type(crystal_t), allocatable, intent(inout) :: vetr(:)
   integer, intent(in)                         :: n

   if (n < 0) return
   if (numcryst(vetr) /= n) then
       if (allocated(vetr))deallocate(vetr)
       if (n > 0) allocate(vetr(n))
   endif

   end subroutine new_crystal_t

!----------------------------------------------------------------------------------------------------

   integer function numdats(dat)
   type(dataset_t), dimension(:), allocatable, intent(in) :: dat
   if (allocated(dat)) then
       numdats = size(dat)
   else
       numdats = 0
   endif
   end function numdats

!----------------------------------------------------------------------------------------------------

   subroutine new_dataset_t(vetr,n)
   type(dataset_t), allocatable, intent(inout) :: vetr(:)
   integer, intent(in)                         :: n

   if (n < 0) return
   if (numdats(vetr) /= n) then
       if (allocated(vetr))deallocate(vetr)
       if (n > 0) allocate(vetr(n))
   endif

   end subroutine new_dataset_t

end module rinterface
